//
//  ATMPPR_Signal.h
//  signals
//
//  Created by atomicpepper on 2/26/17.
//  Copyright © 2017 atomicpepper. All rights reserved.
//

#import <Foundation/Foundation.h>


/*
 * C++ magic...
 */
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

NS_ASSUME_NONNULL_BEGIN
const char *strForSignal(int sigraised);
@interface ATMPPR_Signal : NSObject

// signal handlers should do as little as possible (ideally set a flag and exit).
// To the extent that it is necessary to call external functions from
// a signal handler, only async-signal safe functions should called.


// pass NULL for handlerBlock to "unregister"

+(void)catchSignal:(int)signalNumber signalHandler:(void (^ _Nullable)(int signalRaised))handlerBlock;

// to register for multiple signals at once
+(void)catchSignals:(NSArray<NSNumber*>*)signals signalHandler:(void (^ _Nullable)(int signalRaised))handlerBlock;
+(void)catchSignals:(NSArray<NSNumber*>*)signals queue:(nullable dispatch_queue_t)queue signalHandler:(void (^ _Nullable)(int signalRaised))handlerBlock;


// a try-catch type wrapper for code that may call abort() ...
// this class method will supercede any catchSignal(s) methods shown above (i.e. any handler set there will not be called)
+(void)codeThatMayCallAbort:(void (^)())tryBlock abortResponse:(void (^ _Nullable)())catchBlock;

// signals for BUS error, Segmentation Fault, Floating Point Exception, Illegal Instruction, Bad System call, Abort
+(NSArray<NSNumber*>*)crashSignals;


// Example usage:
/*

 [ATMPPR_Signal catchSignals:ATMPPR_Signal.crashSignals signalHandler:^(int signalRaised) {
    NSLog(@"crashing with signal %d (%s)",signalRaised,strForSignal( signalRaised));
 }];

 
 [ATMPPR_Signal catchSignal:SIGINFO signalHandler:^(int signalRaised) {
     NSLog(@"caught SIGINFO");
    // do something ...
 }];

*/
@end





NS_ASSUME_NONNULL_END
    /*
     * C++ magic...
     */

#ifdef __cplusplus
}
#endif /* __cplusplus */
