//
//  ATMPPR_Signal.m
//  signals
//
//  Created by atomicpepper on 2/26/17.
//  Copyright © 2017 atomicpepper. All rights reserved.
//

#import "ATMPPR_Signal.h"
@interface NSMapTable (ATMPPR_subscripting)
- (nullable id)objectForKeyedSubscript:(id)key;
- (void)setObject:(id)obj forKeyedSubscript:(id)key;
@end

@interface ATMPPR_Signal ()
+(NSMapTable*)mappedHandlers;
@end
const char *strForSignal(int sigraised)
{
    //  return a char string for an signal code
    switch(sigraised)
    {
        case SIGHUP  :    return "HangUP";
        case SIGINT  :    return "INTerrupt";
        case SIGQUIT :    return "QUIT";
        case SIGILL  :    return "ILLegal instruction";
        case SIGABRT :    return "ABoRT";
        case SIGTRAP :    return "trace TRAP";

        case SIGFPE  :    return "Floating Point Exception";
        case SIGKILL :    return "KILL";
        case SIGBUS  :    return "BUS error";
        case SIGSEGV :    return "SEGmentation Violation";
        case SIGSYS  :    return "bad argument to SYStem call";
        case SIGPIPE :    return "write on a PIPE with no one to read it";
        case SIGALRM :    return "ALaRM clock";
        case SIGTERM :    return "software TERMination signal from kill";
        case SIGURG  :    return "URGent condition on IO channel";
        case SIGSTOP :    return "sendable STOP signal not from tty";
        case SIGTSTP :    return "Tty SToP signal";
        case SIGCONT :    return "CONTinue a stopped process";
        case SIGCHLD :    return "to parent on CHiLD stop or exit";
        case SIGTTIN :    return "to readers pgrp upon background TTy read IN";
        case SIGTTOU :    return "like TTin for OUtput";
        case SIGXCPU :    return "eXceeded CPU time limit";
        case SIGXFSZ :    return "eXceeded File SiZe limit";
        case SIGVTALRM:   return "Virtual Time ALaRM";
        case SIGPROF :    return "PROFiling time alarm";
        case SIGUSR1 :    return "USeR 1";
        case SIGUSR2 :    return "USeR 2";

#if  (!defined(_POSIX_C_SOURCE) || defined(_DARWIN_C_SOURCE))
        case SIGIO      :    return "Input/Output possible signal";
        case SIGWINCH   :    return "WINdow size CHanges";
        case SIGINFO    :    return "INFOrmation request";
#endif
            
        default:        return "unknown";
            break;
    }
}

#import <semaphore.h>

sem_t* g_signalSemaphore;
int g_sigRaised;
//jmp_buf g_jmp_buf;
int* getJumpBuffer() {
    static jmp_buf custom_jmp_buf;
    return custom_jmp_buf;
}
static void signalHandler(int sigraised) {
    // signal handlers should do as little as possible (ideally set a flag and exit).
    // To the extent that it is necessary to call external functions from
    // a signal handler, only async-signal safe functions should called.


    g_sigRaised = sigraised;
    sem_post(g_signalSemaphore);

}
static void handlerSIGABRT(int sigraised) {
    // signal handlers should do as little as possible (ideally set a flag and exit).
    // To the extent that it is necessary to call external functions from
    // a signal handler, only async-signal safe functions should called.

    NSLog(@"%@", NSThread.callStackSymbols);

    if(SIGABRT == sigraised) {
        // jump back
        longjmp(getJumpBuffer(), 1);
    }
}

@implementation ATMPPR_Signal
+(NSArray<NSNumber*>*)crashSignals {
    return @[
             @(SIGILL),
             @(SIGABRT),
             @(SIGFPE),
             @(SIGBUS),
             @(SIGSEGV),
             @(SIGSYS),
             ];
}
+(NSArray<NSNumber*>*)stopSignals {
    return @[
             @(SIGINT),
             @(SIGTERM),
             @(SIGHUP),
             @(SIGQUIT),
             @(SIGTSTP),
             @(SIGSTOP),
             ];
    //SIGUSR1
}
+(NSMapTable*)mappedHandlers {
    static NSMapTable* mapTable = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        mapTable = [[NSMapTable alloc] initWithKeyOptions:(NSPointerFunctionsStrongMemory | NSPointerFunctionsObjectPersonality | NSPointerFunctionsCopyIn) valueOptions:(NSPointerFunctionsStrongMemory | NSPointerFunctionsObjectPersonality) capacity:0];
    });
    return mapTable;
}
+(void)codeThatMayCallAbort:(void (^)())tryBlock abortResponse:(void (^)())catchBlock {

    // set the custom SIGABRT handler
    void (*func)(int);
    if ((func = signal(SIGABRT, &handlerSIGABRT)) == SIG_ERR) {
        NSLog(@"Failed to set a custom SIGABRT handler");
    } else {
        // add the handlerBlock
        if(0 == setjmp( getJumpBuffer()) ) {
            tryBlock();
        } else if (catchBlock) {
            catchBlock();
//            NSLog(@"%@", NSThread.callStackSymbols);
        }
    }
    // restore the previous SIGABRT handler
    if ((func = signal(SIGABRT, func)) == SIG_ERR) {
        NSLog(@"Failed to reset SIGABRT handler");
    } else if (&handlerSIGABRT != func) {
        NSLog(@"SIGABRT handler was reset out from underneath!");
    } else {
//        NSLog(@"SIGABRT handler was reset to %p",func);
    }


}
+(void)setSignal:(int)signalNumber queue:(dispatch_queue_t)queue handler:(void (^)(int signalRaised))handlerBlock {

    // check for crashing signal
    if([self.crashSignals containsObject:@(signalNumber)]) {
        // set up the semaphore
        static dispatch_once_t onceToken;
        static dispatch_queue_t semaphoreQueue;
        dispatch_once(&onceToken, ^{
            g_signalSemaphore = sem_open("global_signal_sem", O_CREAT, (S_IRUSR | S_IWUSR), 0);
            semaphoreQueue = dispatch_queue_create("global_signal_semaphore_queue", DISPATCH_QUEUE_SERIAL);
            dispatch_async(semaphoreQueue, ^{
                while(1) {
                    sem_wait(g_signalSemaphore);
                    int sigraised = g_sigRaised;
//                    NSLog(@"%s caught signal %d (%s)",__func__,sigraised,strForSignal( sigraised));
                    NSMapTable* mappedHandlers = ATMPPR_Signal.mappedHandlers;
                    void (^handlerBlock)(int signalRaised) = mappedHandlers[@(sigraised)];
                    if(handlerBlock) {
                        handlerBlock(sigraised);
                    }
                }
            });
        });

        void (*func)(int);
        if ((func = signal(signalNumber, (handlerBlock ? &signalHandler : SIG_DFL))) == SIG_ERR) {
            NSLog(@"Failed to set a signal handler for signal %d \"%s\".\n",signalNumber , strForSignal(signalNumber));
        } else {
            // add the handlerBlock
            
            self.mappedHandlers[@(signalNumber)] = handlerBlock;
        }
    } else if (handlerBlock) {
        dispatch_source_t source = dispatch_source_create(DISPATCH_SOURCE_TYPE_SIGNAL, signalNumber, 0, queue ?: dispatch_get_main_queue() );

        if (source) {
            // ignore signal to prevent app termination
            signal(signalNumber, SIG_IGN);
            // add to the maptable
            self.mappedHandlers[@(signalNumber)] = source;
            dispatch_source_set_event_handler(source, ^{
                handlerBlock(signalNumber);
            });
            dispatch_resume(source);
        }
    } else {
        self.mappedHandlers[@(signalNumber)] = nil;
    }
}
+(void)catchSignal:(int)signalNumber signalHandler:(void (^)(int signalRaised))handlerBlock {
    [self setSignal:signalNumber queue:nil handler:handlerBlock];
}
+(void)catchSignals:(NSArray<NSNumber*>*)signals signalHandler:(void (^)(int signalRaised))handlerBlock {
    [self catchSignals:signals queue:nil signalHandler:handlerBlock];
}
+(void)catchSignals:(NSArray<NSNumber*>*)signals queue:(dispatch_queue_t)queue signalHandler:(void (^)(int signalRaised))handlerBlock {

    for (NSNumber* aSignal in signals) {
        [self setSignal:aSignal.intValue queue:queue handler:handlerBlock];
    }
}
@end
@implementation NSMapTable (ATMPPR_subscripting)

-(id)objectForKeyedSubscript:(id)key {
    return [self objectForKey:key];
}

-(void)setObject:(id)obj forKeyedSubscript:(id)key {
    if (obj != nil) {
        [self setObject:obj forKey:key];
    } else {
        [self removeObjectForKey:key];
    }
}
@end
